from ..exceptions import RelatedVisitFieldError
from .visit_model_mixin import VisitModelMixin


def get_related_visit_model_attr(model_cls) -> str:
    """Returns the field name for the visit model foreign key
    or raises.

    Note: may also be the reverse relation. For example, if this is an
    `appointment` model class, the relation is a reverse relation where
    the attr name, given related to SubjectVisit, has a fld class
    attr in `subjectvisit` and not `subject_visit`.

    If the reverse attr is not found, check value of `related_name` in the
    field class.
    """
    attrs = []
    for fld_cls in model_cls._meta.get_fields():
        if fld_cls.related_model is not None and issubclass(
            fld_cls.related_model, (VisitModelMixin,)
        ):
            attrs.append(fld_cls.name)
            break
    if len(attrs) > 1:
        raise RelatedVisitFieldError(
            f"More than one field is related to a visit model. Got {attrs}."
        )
    elif len(attrs) == 0:
        raise RelatedVisitFieldError(
            f"{model_cls} has no related visit model. "
            f"Expected the related visit model to be an instance "
            "of `VisitModelMixin`."
        )
    return attrs[0]



from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from edc_appointment.models import Appointment
from edc_facility.import_holidays import import_holidays
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_visit_tracking.constants import SCHEDULED
from edc_visit_tracking.form_validators import VisitFormValidator
from edc_visit_tracking.modelform_mixins import VisitTrackingCrfModelFormMixin

from ..helper import Helper
from ..models import BadCrfNoRelatedVisit, CrfOne, SubjectVisit
from ..visit_schedule import visit_schedule1, visit_schedule2


class SubjectVisitForm(forms.ModelForm):
    form_validator_cls = VisitFormValidator

    class Meta:
        model = SubjectVisit
        fields = "__all__"


class TestForm(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.subject_identifier = "12345"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_visit_tracking.subjectvisit"}
        )

    def test_visit_tracking_form_ok(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        form = CrfForm(
            {
                "f1": "1",
                "f2": "2",
                "f3": "3",
                "report_datetime": get_utcnow(),
                "subject_visit": subject_visit.pk,
            }
        )
        self.assertTrue(form.is_valid())
        form.save(commit=True)

    def test_visit_tracking_form_missing_subject_visit(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        form = CrfForm({"f1": "1", "f2": "2", "f3": "3", "report_datetime": get_utcnow()})
        form.is_valid()
        self.assertIn("subject_visit", form._errors)

    def test_visit_tracking_form_missing_subject_visit_fk_raises(self):
        class BadCrfNoRelatedVisitorm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = BadCrfNoRelatedVisit
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        form = BadCrfNoRelatedVisitorm(
            {"f1": "1", "f2": "2", "f3": "3", "report_datetime": get_utcnow()}
        )
        self.assertRaises(ImproperlyConfigured, form.is_valid)

    def test_visit_tracking_form_no_report_datetime(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        form = CrfForm({"f1": "1", "f2": "2", "f3": "3", "subject_visit": subject_visit.pk})
        self.assertFalse(form.is_valid())
        self.assertIn("report_datetime", form._errors)

    def test_visit_tracking_form_report_datetime_validated_against_related_visit(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        for report_datetime in [
            get_utcnow() - relativedelta(months=1),
            get_utcnow() + relativedelta(months=1),
        ]:
            form = CrfForm(
                {
                    "f1": "1",
                    "f2": "2",
                    "f3": "3",
                    "report_datetime": report_datetime,
                    "subject_visit": subject_visit.pk,
                }
            )
            self.assertFalse(form.is_valid())
            self.assertIn("report_datetime", form._errors)

    @override_settings(TIME_ZONE="Africa/Dar_es_Salaam")
    def test_visit_tracking_form_report_datetime_zone(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            reason=SCHEDULED,
            report_datetime=get_utcnow(),
        )
        dte = get_utcnow().astimezone(ZoneInfo("Africa/Dar_es_Salaam"))
        for report_datetime in [
            dte - relativedelta(months=1),
            dte + relativedelta(months=1),
        ]:
            form = CrfForm(
                {
                    "f1": "1",
                    "f2": "2",
                    "f3": "3",
                    "report_datetime": report_datetime,
                    "subject_visit": subject_visit.pk,
                }
            )
            self.assertFalse(form.is_valid())
            self.assertIn("report_datetime", form._errors)

    @override_settings(TIME_ZONE="Africa/Dar_es_Salaam")
    def test_visit_tracking_form_report_datetime_zone2(self):
        class CrfForm(VisitTrackingCrfModelFormMixin, forms.ModelForm):
            report_datetime_field_attr = "report_datetime"

            class Meta:
                model = CrfOne
                fields = "__all__"

        self.helper.consent_and_put_on_schedule()
        appointment = Appointment.objects.all()[0]
        subject_visit = SubjectVisit.objects.create(appointment=appointment, reason=SCHEDULED)
        form = CrfForm(
            {
                "f1": "1",
                "f2": "2",
                "f3": "3",
                "report_datetime": subject_visit.report_datetime,
                "subject_visit": subject_visit.pk,
            }
        )
        form.is_valid()
        self.assertEqual({}, form._errors)
        form.save(commit=True)

        form = CrfForm(
            {
                "f1": "1",
                "f2": "2",
                "f3": "3",
                "report_datetime": subject_visit.report_datetime,
                "subject_visit": subject_visit.pk,
            }
        )
        form.is_valid()
        self.assertIn("subject_visit", form._errors)



from typing import List

from edc_action_item import Action
from edc_action_item.action_with_notification import ActionWithNotification
from edc_adverse_event.constants import DEATH_REPORT_ACTION
from edc_constants.constants import DEAD, HIGH_PRIORITY, YES
from edc_ltfu.constants import LTFU_ACTION

from .constants import VISIT_MISSED_ACTION


class VisitMissedAction(ActionWithNotification):
    name: str = VISIT_MISSED_ACTION
    display_name: str = "Submit Missed Visit"
    notification_display_name: str = " Submit Missed Visit"
    parent_action_names: List[str] = []
    show_link_to_changelist: bool = True
    priority: str = HIGH_PRIORITY
    loss_to_followup_action_name: str = LTFU_ACTION

    reference_model: str = None  # "inte_subject.subjectvisitmissed"
    admin_site_name: str = None  # "inte_prn_admin"

    def get_loss_to_followup_action_name(self) -> str:
        return self.loss_to_followup_action_name

    def get_next_actions(self) -> List[Action]:
        next_actions: List[Action] = []
        next_actions = self.append_to_next_if_required(
            next_actions=next_actions,
            action_name=self.get_loss_to_followup_action_name(),
            required=self.reference_obj.ltfu == YES,
        )
        next_actions = self.append_to_next_if_required(
            next_actions=next_actions,
            action_name=DEATH_REPORT_ACTION,
            required=self.reference_obj.survival_status == DEAD,
        )
        return next_actions


class MissedVisitAction(VisitMissedAction):
    pass



from django.forms import ModelForm
from edc_crf.modelform_mixins import CrfModelFormMixin

from ..form_validators import VisitMissedFormValidator
from .models import SubjectVisitMissed


class SubjectVisitMissedForm(CrfModelFormMixin, ModelForm):
    form_validator_cls = VisitMissedFormValidator

    def validate_against_consent(self):
        pass

    class Meta:
        model = SubjectVisitMissed
        fields = "__all__"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django import forms
from django.conf import settings

from ...crf_date_validator import (
    CrfDateValidator,
    CrfReportDateAllowanceError,
    CrfReportDateBeforeStudyStart,
    CrfReportDateIsFuture,
)
from ..utils import get_related_visit

if TYPE_CHECKING:
    from ...model_mixins import VisitModelMixin


class VisitTrackingCrfModelFormMixinError(Exception):
    pass


class VisitTrackingCrfModelFormMixin:
    """Validates subject visit and report datetime.

    Usually included in the form class declaration with
    `CRfScheduleModelFormMixin`.
    """

    crf_date_validator_cls = CrfDateValidator
    report_datetime_allowance = getattr(settings, "DEFAULT_REPORT_DATETIME_ALLOWANCE", 0)

    def clean(self: Any) -> dict:
        """Triggers a validation error if subject visit is None.

        If subject visit, validate report_datetime.
        """
        if not self.report_datetime_field_attr:
            raise VisitTrackingCrfModelFormMixinError(
                f"Cannot be None. See modelform for {self._meta.model}."
                "Got `report_datetime_field_attr`=None."
            )
        cleaned_data = super().clean()
        self.validate_visit_tracking()
        return cleaned_data

    @property
    def subject_identifier(self):
        """Overridden"""
        return self.related_visit.subject_identifier

    @property
    def related_visit(self) -> VisitModelMixin | None:
        return get_related_visit(self, related_visit_model_attr=self.related_visit_model_attr)

    @property
    def related_visit_model_attr(self) -> str:
        try:
            return self._meta.model.related_visit_model_attr()
        except AttributeError:
            raise VisitTrackingCrfModelFormMixinError(
                "Expected method `related_visit_model_attr`. Is this a CRF? "
                f"See model {self._meta.model}"
            )

    def validate_visit_tracking(self: Any) -> None:
        # trigger a validation error if visit field is None
        # no comment needed since django will catch it as
        # a required field.
        if not self.related_visit:
            if self.related_visit_model_attr in self.cleaned_data:
                raise forms.ValidationError({self.related_visit_model_attr: ""})
            else:
                raise forms.ValidationError(
                    f"Field `{self.related_visit_model_attr}` is required (1)."
                )
        elif self.cleaned_data.get(self.report_datetime_field_attr):
            try:
                self.crf_date_validator_cls(
                    report_datetime_allowance=self.report_datetime_allowance,
                    report_datetime=self.cleaned_data.get(self.report_datetime_field_attr),
                    visit_report_datetime=self.related_visit.report_datetime,
                )
            except (
                CrfReportDateAllowanceError,
                CrfReportDateBeforeStudyStart,
                CrfReportDateIsFuture,
            ) as e:
                raise forms.ValidationError({self.report_datetime_field_attr: str(e)})

    # @property
    # def report_datetime(self) -> datetime:
    #     """Overridden. Returns the report_datetime as UTC from directly from
    #     cleaned_data or via the related_visit, if it exists.
    #     """
    #     report_datetime = self.cleaned_data.get(self.report_datetime_field_attr)
    #     if self.related_visit and not report_datetime:
    #         report_datetime = getattr(self.related_visit, self.report_datetime_field_attr)
    #     if report_datetime:
    #         report_datetime = report_datetime.astimezone(ZoneInfo("UTC"))
    #     return report_datetime



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from django.apps import apps as django_apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

if TYPE_CHECKING:
    from edc_list_data.model_mixins import ListModelMixin

    from .model_mixins import SubjectVisitMissedModelMixin, VisitModelMixin

    VisitModel = TypeVar("VisitModel", bound=VisitModelMixin)
    VisitMissedModel = TypeVar("VisitMissedModel", bound=SubjectVisitMissedModelMixin)
    ListModel = TypeVar("ListModel", bound=ListModelMixin)


def get_related_visit_model() -> str:
    return settings.SUBJECT_VISIT_MODEL


def get_related_visit_model_cls() -> VisitModel:
    return django_apps.get_model(settings.SUBJECT_VISIT_MODEL)


def get_subject_visit_model() -> str:
    return get_related_visit_model()


def get_subject_visit_model_cls() -> VisitModel:
    return get_related_visit_model_cls()


def get_subject_visit_missed_reasons_model() -> ListModel:
    error_msg = (
        "Settings attribute `SUBJECT_VISIT_MISSED_REASONS_MODEL` not set. "
        "Update settings. For example, `SUBJECT_VISIT_MISSED_REASONS_MODEL"
        "=meta_lists.subjectvisitmissedreasons`. "
        "See also `SubjectVisitMissedModelMixin`."
    )
    try:
        model = settings.SUBJECT_VISIT_MISSED_REASONS_MODEL
    except AttributeError as e:
        raise ImproperlyConfigured(f"{error_msg} Got {e}.")
    else:
        if not model:
            raise ImproperlyConfigured(f"{error_msg} Got None.")
    return model


def get_subject_visit_missed_model() -> str:
    error_msg = (
        "Settings attribute `SUBJECT_VISIT_MISSED_MODEL` not set. Update settings. "
        "For example, `SUBJECT_VISIT_MISSED_MODEL=meta_subject.subjectvisitmissed`. "
        "See also `SubjectVisitMissedModelMixin`."
    )
    try:
        model = settings.SUBJECT_VISIT_MISSED_MODEL
    except AttributeError as e:
        raise ImproperlyConfigured(f"{error_msg} Got {e}.")
    else:
        if not model:
            raise ImproperlyConfigured(f"{error_msg} Got None.")
    return model


def get_subject_visit_missed_model_cls() -> VisitMissedModel:
    return django_apps.get_model(get_subject_visit_missed_model())



from edc_constants.constants import ALIVE, NO, OTHER, UNKNOWN, YES
from edc_form_validators import FormValidator


class VisitMissedFormValidator(FormValidator):
    def clean(self) -> None:
        self.applicable_if(YES, field="contact_attempted", field_applicable="contact_made")
        self.required_if(
            YES,
            field="contact_attempted",
            field_required="contact_attempts_count",
            field_required_evaluate_as_int=True,
        )
        contact_attempts_count = (
            0
            if self.cleaned_data.get("contact_attempts_count") is None
            else self.cleaned_data.get("contact_attempts_count")
        )
        cond = self.cleaned_data.get("contact_made") == NO and contact_attempts_count < 3
        self.required_if_true(
            cond,
            field_required="contact_attempts_explained",
        )

        self.required_if(YES, field="contact_attempted", field_required="contact_last_date")

        self.m2m_required_if(YES, field="contact_made", m2m_field="missed_reasons")

        self.m2m_other_specify(
            OTHER, m2m_field="missed_reasons", field_other="missed_reasons_other"
        )
        self.applicable_if(ALIVE, UNKNOWN, field="survival_status", field_applicable="ltfu")



from django.db import models
from django.db.models.deletion import PROTECT
from edc_constants.constants import HOSPITALIZED, OTHER
from edc_crf.model_mixins import CrfInlineModelMixin, CrfModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_lab.model_mixins import RequisitionModelMixin
from edc_list_data.model_mixins import ListModelMixin
from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin
from edc_model.models import BaseUuidModel
from edc_offstudy.model_mixins import OffstudyModelMixin
from edc_reference.model_mixins import ReferenceModelMixin
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import OffScheduleModelMixin, OnScheduleModelMixin

from edc_visit_tracking.model_mixins import SubjectVisitMissedModelMixin
from edc_visit_tracking.models import SubjectVisitMissedReasons

from ..choices import VISIT_INFO_SOURCE, VISIT_REASON, VISIT_REASON_MISSED
from ..model_mixins import VisitModelMixin, VisitTrackingCrfModelMixin

list_data = {
    "edc_visit_tracking.subjectvisitmissedreasons": [
        ("forgot", "Forgot / Can’t remember being told about appointment"),
        ("family_emergency", "Family emergency (e.g. funeral) and was away"),
        ("travelling", "Away travelling/visiting"),
        ("working_schooling", "Away working/schooling"),
        ("too_sick", "Too sick or weak to come to the centre"),
        ("lack_of_transport", "Transportation difficulty"),
        (HOSPITALIZED, "Hospitalized"),
        (OTHER, "Other reason (specify below)"),
    ],
}


class SubjectConsent(
    NonUniqueSubjectIdentifierFieldMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    BaseUuidModel,
):
    consent_datetime = models.DateTimeField(default=get_utcnow)

    report_datetime = models.DateTimeField(default=get_utcnow)

    class Meta(BaseUuidModel.Meta):
        pass


class OnScheduleOne(OnScheduleModelMixin, BaseUuidModel):
    class Meta(BaseUuidModel.Meta):
        pass


class OnScheduleTwo(OnScheduleModelMixin, BaseUuidModel):
    class Meta(BaseUuidModel.Meta):
        pass


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):
    class Meta(BaseUuidModel.Meta):
        pass


class OffScheduleOne(OffScheduleModelMixin, BaseUuidModel):
    class Meta(BaseUuidModel.Meta):
        pass


class SubjectOffstudy(OffstudyModelMixin, BaseUuidModel):
    class Meta(OffstudyModelMixin.Meta):
        pass


class SubjectVisit(
    VisitModelMixin,
    ReferenceModelMixin,
    CreatesMetadataModelMixin,
    SiteModelMixin,
    BaseUuidModel,
):
    reason = models.CharField(max_length=25, choices=VISIT_REASON)

    reason_missed = models.CharField(
        verbose_name="If 'missed', provide the reason for the missed visit",
        max_length=35,
        choices=VISIT_REASON_MISSED,
        blank=True,
        null=True,
    )

    info_source = models.CharField(
        verbose_name="What is the main source of this information?",
        max_length=25,
        choices=VISIT_INFO_SOURCE,
    )

    class Meta(VisitModelMixin.Meta, BaseUuidModel.Meta):
        pass


class SubjectRequisition(RequisitionModelMixin, BaseUuidModel):
    def update_reference_on_save(self):
        pass

    class Meta(RequisitionModelMixin.Meta):
        pass


class CrfOne(VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class CrfTwo(VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class CrfThree(VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class CrfFour(VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class CrfFive(VisitTrackingCrfModelMixin, BaseUuidModel):
    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class BadCrfNoRelatedVisit(VisitTrackingCrfModelMixin, BaseUuidModel):
    subject_visit = None

    f1 = models.CharField(max_length=50, null=True)

    f2 = models.CharField(max_length=50, null=True)

    f3 = models.CharField(max_length=50, null=True)

    class Meta(BaseUuidModel.Meta):
        pass


class OtherModel(BaseUuidModel):
    f1 = models.CharField(max_length=10, default="erik")

    class Meta(BaseUuidModel.Meta):
        pass


class CrfOneInline(CrfInlineModelMixin, BaseUuidModel):
    crf_one = models.ForeignKey(CrfOne, on_delete=PROTECT)

    other_model = models.ForeignKey(OtherModel, on_delete=PROTECT)

    f1 = models.CharField(max_length=10, default="erik")

    def natural_key(self) -> tuple:
        return tuple()

    class Meta(CrfInlineModelMixin.Meta):
        crf_inline_parent = "crf_one"


class BadCrfOneInline(CrfInlineModelMixin, BaseUuidModel):
    """A model class missing _meta.crf_inline_parent."""

    crf_one = models.ForeignKey(CrfOne, on_delete=PROTECT)

    other_model = models.ForeignKey(OtherModel, on_delete=PROTECT)

    f1 = models.CharField(max_length=10, default="erik")

    def natural_key(self) -> tuple:
        return tuple()

    class Meta:
        pass


class BadCrfOneInline2(CrfInlineModelMixin, BaseUuidModel):
    crf_one = models.ForeignKey(CrfOne, on_delete=PROTECT)

    other_model = models.ForeignKey(OtherModel, on_delete=PROTECT)

    f1 = models.CharField(max_length=10, default="erik")

    def natural_key(self) -> tuple:
        return tuple()

    class Meta(CrfInlineModelMixin.Meta):
        crf_inline_parent = None


class CustomSubjectVisitMissedReasons(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Subject Visit Missed Reason"
        verbose_name_plural = "Subject Visit Missed Reasons"


class SubjectVisitMissed(
    CrfModelMixin,
    SubjectVisitMissedModelMixin,
    BaseUuidModel,
):
    missed_reasons = models.ManyToManyField(SubjectVisitMissedReasons, blank=True)

    class Meta(VisitTrackingCrfModelMixin.Meta):
        pass
